import ast
from typing import Any, Dict, List, Optional

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable


class CountMethod(IBuiltinMethod):

    def __init__(self, args: Dict[str, Variable] = None, defaults: List[ast.AST] = None):
        from boa3.model.type.type import Type
        identifier = 'count'
        super().__init__(identifier, args, defaults=defaults, return_type=Type.int)

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type

        if self._arg_self.type is Type.str:     # CountStrMethod
            return '-{0}_{1}'.format(self._identifier, Type.str.identifier)

        if Type.sequence.is_type_of(self._arg_self.type):

            from boa3.model.type.annotation.uniontype import UnionType
            from boa3.model.type.primitive.bytestype import BytesType
            from boa3.model.type.primitive.inttype import IntType
            from boa3.model.type.primitive.strtype import StrType

            if isinstance(self._arg_self.type.value_type, (BytesType, IntType, StrType)) or\
                    ((isinstance(self._arg_self.type.value_type, UnionType) and
                      all(isinstance(type_, (BytesType, IntType, StrType)) for type_ in self._arg_self.type.value_type.union_types))):    # CountSequencePrimitiveValueMethod
                return '-{0}_{1}'.format(self._identifier, 'primitive')

            elif Type.mutableSequence.is_type_of(self._arg_self.type.value_type):    # CountSequenceGenericValueMethod
                return self._identifier

        return self._identifier

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def _arg_value(self) -> Variable:
        return self.args['value']

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, list):
            value = [value]

        from boa3.model.type.collection.sequence.mutable.listtype import ListType
        from boa3.model.type.collection.sequence.tupletype import TupleType
        from boa3.model.type.collection.sequence.rangetype import RangeType
        if len(value) > 1 and isinstance(value[0], (ListType, TupleType, RangeType)):

            from boa3.model.type.primitive.bytestype import BytesType
            from boa3.model.type.primitive.inttype import IntType
            from boa3.model.type.primitive.strtype import StrType
            from boa3.model.type.annotation.uniontype import UnionType

            if isinstance(value[0].value_type, (BytesType, IntType, StrType)) or\
                    ((isinstance(value[0].value_type, UnionType) and
                      all(isinstance(type_, (BytesType, IntType, StrType)) for type_ in value[0].value_type.union_types))):

                from boa3.model.builtin.classmethod.countsequenceprimitivemethod import CountSequencePrimitiveMethod
                return CountSequencePrimitiveMethod(value[0], value[1])

            from boa3.model.builtin.classmethod.countsequencegenericmethod import CountSequenceGenericMethod
            return CountSequenceGenericMethod(value[0], value[1])

        from boa3.model.type.type import Type
        if len(value) > 0 and Type.str.is_type_of(value[0]):
            from boa3.model.builtin.builtin import Builtin
            return Builtin.CountStr

        return super().build(value)



import ast
from typing import Any, Dict, List, Optional, Tuple, Union

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class RangeMethod(IBuiltinMethod):

    def __init__(self, values_types: List[IType] = None):
        from boa3.model.type.type import Type
        identifier = 'range'
        if not isinstance(values_types, list) or len(values_types) < 1:
            stop = Variable(Type.none)
            start = Variable(Type.none)
            step = Variable(Type.none)
        else:
            stop = Variable(Type.int)
            start = Variable(Type.int)
            step = Variable(Type.int)

        args: Dict[str, Variable] = {
            'stop': stop,
            'start': start,
            'step': step
        }
        start_default = ast.parse("{0}".format(Type.int.default_value)
                                  ).body[0].value
        step_default = ast.parse("1").body[0].value
        super().__init__(identifier, args, defaults=[start_default, step_default], return_type=Type.range)

    @property
    def _arg_stop(self) -> Variable:
        return self.args['stop']

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if self._arg_stop.type is Type.none:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_stop.type._identifier)

    @property
    def requires_reordering(self) -> bool:
        return True

    def reorder(self, arguments: list):
        if len(arguments) > 1:
            # swap start and stop default positions
            arguments[0], arguments[1] = arguments[1], arguments[0]

    def validate_parameters(self, *params: Union[IExpression, IType]) -> bool:
        if len(params) < 1 or len(params) > 3:
            return False

        if any(not isinstance(param, (IExpression, IType)) for param in params):
            return False

        params_type: List[IType] = [param if isinstance(param, IType) else param.type for param in params]
        from boa3.model.type.type import Type
        return all(param is Type.int for param in params_type)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        from boa3.neo.vm.type.String import String
        range_error_msg = String('range() arg 3 must not be zero').to_bytes()
        return [
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIGN, b''),
            (Opcode.JMPIF, Integer(5 + len(range_error_msg)).to_byte_array(signed=True)),
            (Opcode.PUSHDATA1, Integer(len(range_error_msg)).to_byte_array(signed=True) + range_error_msg),
            (Opcode.THROW, b''),
            (Opcode.NEWARRAY0, b''),
            (Opcode.REVERSE4, b''),
            (Opcode.SWAP, b''),
            (Opcode.JMP, Integer(8).to_byte_array(signed=True, min_length=1)),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.APPEND, b''),
            (Opcode.OVER, b''),
            (Opcode.ADD, b''),
            (Opcode.DUP, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIGN, b''),
            (Opcode.PUSH0, b''),
            (Opcode.JMPGT, Integer(5).to_byte_array(signed=True, min_length=1)),
            (Opcode.GT, b''),
            (Opcode.JMP, Integer(3).to_byte_array(signed=True, min_length=1)),
            (Opcode.LT, b''),
            (Opcode.JMPIF, Integer(-19).to_byte_array(signed=True, min_length=1)),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b'')
        ]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and self.validate_parameters(*value):
            return RangeMethod(value)
        return super().build(value)



import ast
from typing import Dict, List, Tuple

from boa3.model.builtin.method.listmethod import ListMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class ListGenericMethod(ListMethod):

    def __init__(self, value: IType = None):
        from boa3.model.type.type import Type

        if value is None:
            value = Type.any

        args: Dict[str, Variable] = {
            'value': Variable(value),
        }

        value_default = ast.parse("{0}".format(Type.sequence.default_value)
                                  ).body[0].value

        return_value = Type.any if value is Type.any else []

        if return_value is not Type.any:
            for type_ in value.union_types:
                if type_ is Type.str:
                    return_value.append(Type.str)
                elif type_ is Type.bytes:
                    return_value.append(Type.int)
                elif Type.sequence.is_type_of(type_):
                    return_value.append(type_.value_type)
                elif Type.mapping.is_type_of(type_):
                    return_value.append(type_.key_type)

            if Type.str in value.union_types and Type.bytes in value.union_types:
                return_value.remove(Type.str)

        return_type = Type.list.build_collection(return_value)

        super().__init__(args, return_type, [value_default])

    @property
    def prepare_for_packing(self) -> List[Tuple[Opcode, bytes]]:

        if self._prepare_for_packing is None:
            from boa3.compiler.codegenerator import get_bytes_count
            from boa3.neo.vm.type.StackItem import StackItemType
            from boa3.model.type.annotation.uniontype import UnionType
            from boa3.model.type.type import Type
            from boa3.neo.vm.type.Integer import Integer
            from boa3.neo.vm.type.String import String

            jmp_place_holder = (Opcode.JMP, b'\x01')
            error_message = String('Invalid value given, it should be an iterable').to_bytes()

            verify_is_bytes_string = [
                (Opcode.DUP, b''),
                (Opcode.ISTYPE, StackItemType.ByteString),
                jmp_place_holder
            ]

            from boa3.model.builtin.method import ListBytesStringMethod
            if isinstance(self._arg_value, UnionType) and \
                    Type.str in self._arg_value.union_types and Type.bytes not in self._arg_value.union_types:
                pre_packing_bytes_string = ListBytesStringMethod(Type.str).prepare_for_packing
            else:
                pre_packing_bytes_string = ListBytesStringMethod().prepare_for_packing

            verify_is_sequence = [
                (Opcode.DUP, b''),
                (Opcode.ISTYPE, StackItemType.Array),
                jmp_place_holder
            ]

            from boa3.model.builtin.method import ListSequenceMethod
            pre_packing_sequence = ListSequenceMethod().prepare_for_packing

            verify_is_mapping = [
                (Opcode.DUP, b''),
                (Opcode.ISTYPE, StackItemType.Map),
                jmp_place_holder
            ]

            from boa3.model.builtin.method import ListMappingMethod
            pre_packing_mapping = ListMappingMethod().prepare_for_packing

            invalid_value = [
                (Opcode.PUSHDATA1, Integer(len(error_message)).to_byte_array(signed=True) + error_message),
                (Opcode.THROW, b'')
            ]

            num_jmp_code = get_bytes_count(invalid_value)
            jmp_to_end_from_mapping = OpcodeHelper.get_jump_and_data(Opcode.JMP, num_jmp_code, True)
            pre_packing_mapping.append(jmp_to_end_from_mapping)

            num_jmp_code = get_bytes_count(pre_packing_mapping)
            jmp_prepare_mapping = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)
            verify_is_mapping.append(jmp_prepare_mapping)

            num_jmp_code = get_bytes_count(invalid_value + pre_packing_mapping + verify_is_mapping)
            jmp_to_end_from_sequence = OpcodeHelper.get_jump_and_data(Opcode.JMP, num_jmp_code, True)
            pre_packing_sequence.append(jmp_to_end_from_sequence)

            num_jmp_code = get_bytes_count(pre_packing_sequence)
            jmp_prepare_sequence = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)
            verify_is_sequence.append(jmp_prepare_sequence)

            num_jmp_code = get_bytes_count(invalid_value + pre_packing_mapping + verify_is_mapping +
                                           pre_packing_sequence + verify_is_sequence)
            jmp_to_end_from_bytes_string = OpcodeHelper.get_jump_and_data(Opcode.JMP, num_jmp_code, True)
            pre_packing_bytes_string.append(jmp_to_end_from_bytes_string)

            num_jmp_code = get_bytes_count(pre_packing_bytes_string)
            jmp_prepare_bytes_string = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)
            verify_is_bytes_string.append(jmp_prepare_bytes_string)

            self._prepare_for_packing = (
                verify_is_bytes_string +
                pre_packing_bytes_string +
                verify_is_sequence +
                pre_packing_sequence +
                verify_is_mapping +
                pre_packing_mapping +
                invalid_value
            )

        return super().prepare_for_packing



from typing import Dict

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.nativecontract import LedgerMethod
from boa3.model.variable import Variable


class GetCurrentIndexMethod(LedgerMethod):
    def __init__(self):
        identifier = 'get_current_index'
        syscall = 'currentIndex'
        args: Dict[str, Variable] = {}
        from boa3.model.type.type import Type
        super().__init__(identifier, syscall, args, return_type=Type.int)


class CurrentIndexProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'current_index'
        getter = GetCurrentIndexMethod()
        super().__init__(identifier, getter)



from __future__ import annotations

import hashlib
from typing import List, Tuple

from boa3.neo3 import contracts
from boa3.neo3.core import Size as s, serialization, types, utils


class Version(serialization.ISerializable):
    """
    Represents the version number of an assembly
    """

    def __init__(self, major: int = 0, minor: int = 0, build: int = 0, revision: int = 0):
        """
        Args:
            major: non interchangeable assembly.
            minor: significant enhancements with backwards compatibility.
            build: recompilation of the same source with possible other compiler or on other platform.
            revision: fully interchangeable. Can be used for example for security fixes.
        """
        if major < 0 or minor < 0 or build < 0 or revision < 0:
            raise ValueError("Negative version numbers are not allowed")

        if major > 255 or minor > 255 or build > 255 or revision > 255:
            raise ValueError("Version numbers cannot exceed 255")

        self.major = major
        self.minor = minor
        self.build = build
        self.revision = revision

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.major == other.major
                and self.minor == other.minor
                and self.build == other.build
                and self.revision == other.revision)

    def __len__(self):
        return s.uint64 + s.uint64 + s.uint64 + s.uint64

    def __str__(self):
        return "{0}.{1}.{2}.{3}".format(self.major,
                                        self.minor,
                                        self.build,
                                        self.revision)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        version_str = str(self)
        from boa3.neo.vm.type.String import String
        version_bytes = String(version_str).to_bytes() + bytes(s.uint64 * 4 - len(version_str))
        writer.write_bytes(version_bytes)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        version_str = reader.read_bytes(s.uint64 * 4).decode('utf-8')

        import re
        version_str = re.sub(r'\x00+', '', version_str)

        version_split = version_str.split('.')
        while len(version_split) < 4:
            version_split.append('0')

        major, minor, build, revision = version_split
        self.major = int(major)
        self.minor = int(minor)
        self.build = int(build)
        self.revision = int(revision)

    @classmethod
    def _parse_component(self, c: str) -> Tuple[bool, int]:
        try:
            r = int(c)
        except ValueError:
            return False, -1
        if r < 0:
            return False, -1
        if r > 255:
            return False, -1
        return True, r

    @classmethod
    def from_string(cls, input: str) -> Version:
        """
        Parse an instance out of a string.

        Args:
            input: string representing a version number following the format `Major.Minor[.build[.revision]]`.
            Each version part must fit in the range >= 0 <= 255.

        Raises:
            ValueError: if the input cannot be successfully parsed.
        """
        parts = input.split('.')
        if len(parts) < 2 or len(parts) > 4:
            raise ValueError(f"Cannot parse version from: {input}")

        success, major = Version._parse_component(parts[0])
        if not success:
            raise ValueError(f"Cannot parse major field from: {parts[0]}")
        success, minor = Version._parse_component(parts[1])
        if not success:
            raise ValueError(f"Cannot parse minor field from: {parts[1]}")

        if len(parts) > 2:
            success, build = Version._parse_component(parts[2])
            if not success:
                raise ValueError(f"Cannot parse build field from: {parts[2]}")
        else:
            build = 0

        if len(parts) > 3:
            success, revision = Version._parse_component(parts[3])
            if not success:
                raise ValueError(f"Cannot parse revision field from: {parts[3]}")
        else:
            revision = 0
        return cls(major, minor, build, revision)


class NEF(serialization.ISerializable):
    def __init__(self,
                 compiler_name: str = None,
                 script: bytes = None,
                 tokens: List[MethodToken] = None,
                 source: str = None,
                 _magic: int = 0x3346454E):
        self.magic = _magic
        if compiler_name is None:
            self.compiler = 'unknown'
        else:
            self.compiler = compiler_name[:64]
        self.source = source if source else ""
        self.script = script if script else b''
        self._checksum = 0
        self.tokens = [] if tokens is None else tokens
        # this is intentional, because NEO computes the initial checksum by serializing itself while checksum is 0
        self._checksum = self.compute_checksum()

    def __len__(self):
        return (
            s.uint32  # magic
            + 64  # compiler
            + utils.get_var_size(self.source)
            + 1  # reserved
            + utils.get_var_size(self.tokens)
            + 2  # reserved
            + s.uint32  # checksum
            + utils.get_var_size(self.script)
        )

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.magic == other.magic
                and self.compiler == other.compiler
                and self.source == other.source
                and self.script == other.script
                and self.tokens == other.tokens
                and self.checksum == other.checksum
                )

    @property
    def checksum(self) -> int:
        if self._checksum == 0:
            self._checksum = self.compute_checksum()
        return self._checksum

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint32(self.magic)
        writer.write_bytes(self.compiler.encode('utf-8').ljust(64, b'\x00'))
        writer.write_var_string(self.source)
        writer.write_bytes(b'\x00')
        writer.write_serializable_list(self.tokens)
        writer.write_bytes(b'\x00\x00')
        writer.write_var_bytes(self.script)
        writer.write_uint32(self._checksum)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        if reader.read_uint32() != self.magic:
            raise ValueError("Deserialization error - Incorrect magic")
        self.compiler = reader.read_bytes(64).decode('utf-8').rstrip(b'\x00'.decode())
        self.source = reader.read_var_string(256)
        if reader.read_uint8() != 0:
            raise ValueError("Reserved bytes must be 0")
        self.tokens = reader.read_serializable_list(MethodToken)
        if reader.read_uint16() != 0:
            raise ValueError("Reserved bytes must be 0")

        self.script = reader.read_var_bytes(max=512 * 1024)
        if len(self.script) == 0:
            raise ValueError("Deserialization error - Script can't be empty")

        checksum = int.from_bytes(reader.read_bytes(4), 'little')
        if checksum != self.compute_checksum():
            raise ValueError("Deserialization error - Invalid checksum")
        else:
            self._checksum = checksum

    def compute_checksum(self) -> int:
        return int.from_bytes(
            hashlib.sha256(hashlib.sha256(self.to_array()[:-4]).digest()).digest()[:s.uint32],
            'little'
        )

    @classmethod
    def _serializable_init(cls):
        c = cls()
        c._checksum = 0
        return c


class MethodToken(serialization.ISerializable):
    def __init__(self,
                 hash: types.UInt160,
                 method: str,
                 parameters_count: int,
                 has_return_value: bool,
                 call_flags: contracts.CallFlags):
        self.hash = hash
        self.method = method
        self.parameters_count = parameters_count
        self.has_return_value = has_return_value
        self.call_flags = call_flags

    def __len__(self):
        return s.uint160 + utils.get_var_size(self.method) + s.uint16 + s.uint8 + s.uint8

    def serialize(self, writer: serialization.BinaryWriter):
        writer.write_serializable(self.hash)
        writer.write_var_string(self.method)
        writer.write_uint16(self.parameters_count)
        writer.write_uint8(self.has_return_value)
        writer.write_uint8(self.call_flags.value)

    def deserialize(self, reader: serialization.BinaryReader):
        self.hash = reader.read_serializable(types.UInt160)
        self.method = reader.read_var_string(32)
        self.parameters_count = reader.read_uint16()
        self.has_return_value = bool(reader.read_uint8())
        self.call_flags = contracts.CallFlags(reader.read_uint8())

    @classmethod
    def _serializable_init(cls):
        return cls(types.UInt160.zero(), "", 0, False, contracts.CallFlags.NONE)



from __future__ import annotations

from typing import Any


class Iterator:
    """
    The iterator for smart contracts.
    """

    @property
    def value(self) -> Any:
        """
        Gets the element in the collection at the current position of the iterator.

        :return: the element in the collection at the current position of the iterator
        :rtype: Any
        """
        return None

    def next(self) -> bool:
        """
        Advances the iterator to the next element of the collection.

        :return: true if it advanced, false if there isn't a next element
        :rtype: bool
        """
        pass



from typing import Any, Dict

from boa3.model.type.annotation.optionaltype import OptionalType
from boa3.model.type.annotation.uniontype import UnionType
from boa3.model.type.anytype import anyType
from boa3.model.type.baseexceptiontype import BaseExceptionType
from boa3.model.type.collection.genericcollectiontype import GenericCollectionType
from boa3.model.type.collection.mapping.genericmappingtype import GenericMappingType
from boa3.model.type.collection.mapping.mutable.dicttype import DictType
from boa3.model.type.collection.sequence.genericsequencetype import GenericSequenceType
from boa3.model.type.collection.sequence.mutable.genericmutablesequencetype import GenericMutableSequenceType
from boa3.model.type.collection.sequence.mutable.listtype import ListType
from boa3.model.type.collection.sequence.rangetype import RangeType
from boa3.model.type.collection.sequence.reversedtype import ReversedType
from boa3.model.type.collection.sequence.tupletype import TupleType
from boa3.model.type.itype import IType
from boa3.model.type.primitive.booltype import BoolType
from boa3.model.type.primitive.bytearraytype import ByteArrayType
from boa3.model.type.primitive.bytestype import BytesType
from boa3.model.type.primitive.inttype import IntType
from boa3.model.type.primitive.nonetype import noneType
from boa3.model.type.primitive.strtype import StrType


class Type:
    @classmethod
    def builtin_types(cls) -> Dict[str, IType]:
        """
        Gets a dictionary that maps each type with its name

        :return: A dictionary that maps each Python builtin type representation with its name
        """
        builtin_types = [
            Type.int,
            Type.bool,
            Type.str,
            Type.list,
            Type.tuple,
            Type.dict,
            Type.range,
            Type.reversed,
            Type.bytes,
            Type.bytearray,
            Type.none
        ]
        return {tpe._identifier: tpe for tpe in builtin_types if isinstance(tpe, IType)}

    @classmethod
    def all_types(cls) -> Dict[str, IType]:
        return {tpe._identifier: tpe for tpe in vars(cls).values() if isinstance(tpe, IType)}

    @classmethod
    def get_type(cls, value: Any) -> IType:
        """
        Returns the type of the given value.

        :param value: value to get the type
        :return: Returns the type of the value. `Type.none` by default.
        """
        val: IType = None
        for type in vars(cls).values():
            if isinstance(type, IType) and type.is_type_of(value):
                val = type.build(value)
                break

        if val is not None:
            return val
        return cls.none

    @classmethod
    def get_generic_type(cls, *types: IType) -> IType:
        """
        Returns the common generic type between the given values.

        :param types: list of type to be compared
        :return: Returns the common generic type of the values if exist. `Type.any` otherwise.
        """
        generic = cls.any
        if len(types) > 0:
            generic_types = [cls.mutableSequence, cls.mapping, cls.sequence]
            generic_types.append(generic)  # any type must be the last value in the list

            generic = types[0]
            for tpe in types[1:]:
                if generic == tpe or generic.is_type_of(tpe):
                    continue
                if tpe.is_type_of(generic):
                    generic = tpe
                    continue

                for gen in generic_types:
                    if gen.is_type_of(generic) and gen.is_type_of(tpe):
                        generic = gen
                        break

                if generic is Type.any:
                    break
        if generic is Type.any and Type.any not in types:
            generic = Type.union.build(types)
        return generic

    # Builtin Types
    int = IntType()
    bool = BoolType()
    str = StrType()
    none = noneType
    bytes = BytesType()
    bytearray = ByteArrayType()
    tuple = TupleType()
    list = ListType()
    dict = DictType()
    range = RangeType(int)
    reversed = ReversedType()

    # Generic types
    sequence = GenericSequenceType()
    mutableSequence = GenericMutableSequenceType()
    mapping = GenericMappingType()
    collection = GenericCollectionType()
    exception = BaseExceptionType()

    # Annotation types
    union = UnionType()
    optional = OptionalType()
    any = anyType



import ast
from typing import Any, Dict, Iterable, List, Sized, Tuple

from boa3.model import set_internal_call
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class StorageGetMethod(InteropMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.builtin.interop.storage.storagecontext.storagecontexttype import StorageContextType
        from boa3.model.type.primitive.bytestringtype import ByteStringType

        identifier = 'get'
        syscall = 'System.Storage.Get'
        context_type = StorageContextType.build()
        byte_string_type = ByteStringType.build()

        args: Dict[str, Variable] = {'key': Variable(byte_string_type),
                                     'context': Variable(context_type)}

        from boa3.model.builtin.interop.storage.storagegetcontextmethod import StorageGetContextMethod
        default_id = StorageGetContextMethod(context_type).identifier
        context_default = set_internal_call(ast.parse("{0}()".format(default_id)
                                                      ).body[0].value)
        super().__init__(identifier, syscall, args, defaults=[context_default], return_type=Type.bytes)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.type.type import Type
        from boa3.neo.vm.type.Integer import Integer

        opcodes = super()._opcode
        opcodes.extend([
            (Opcode.DUP, b''),
            (Opcode.ISNULL, b''),
            (Opcode.JMPIFNOT, Integer(7).to_byte_array(signed=True, min_length=1)),
            (Opcode.DROP, b''),
            (Opcode.PUSHDATA1, b'\x00'),
            (Opcode.CONVERT, Type.bytes.stack_item),
        ])
        return opcodes

    @property
    def generation_order(self) -> List[int]:
        """
        Gets the indexes order that need to be used during code generation.
        If the order for generation is the same as inputted in code, returns reversed(range(0,len_args))

        :return: Index order for code generation
        """
        indexes = super().generation_order
        context_index = list(self.args).index('context')

        if indexes[-1] != context_index:
            # context must be the last generated argument
            indexes.remove(context_index)
            indexes.append(context_index)

        return indexes

    @property
    def key_arg(self) -> Variable:
        return self.args['key']

    def build(self, value: Any) -> IBuiltinMethod:
        exp: List[IExpression] = []
        if isinstance(value, Sized):
            if len(value) > 1 or not isinstance(value, Iterable):
                return self
            exp = [exp if isinstance(exp, IExpression) else Variable(exp)
                   for exp in value if isinstance(exp, (IExpression, IType))]

        elif isinstance(exp, (IExpression, IType)):
            exp = [value if isinstance(value, IExpression) else Variable(value)]
        else:
            return self

        if not self.validate_parameters(*exp):
            return self

        method = self
        key_type: IType = exp[0].type
        if not method.key_arg.type.is_type_of(key_type):
            method = StorageGetMethod()
            method.args['key'] = Variable(key_type)
        return method



from typing import Dict

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class ScriptContainerMethod(InteropMethod):
    def __init__(self):
        from boa3.model.type.type import Type
        identifier = '-get_script_container'
        syscall = 'System.Runtime.GetScriptContainer'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=Type.any)


class ScriptContainerProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'script_container'
        getter = ScriptContainerMethod()
        super().__init__(identifier, getter)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError, CompilerWarning
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.StackItem import StackItemType
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestBytes(BoaTest):
    default_folder: str = 'test_sc/bytes_test'

    SUBSEQUENCE_NOT_FOUND_MSG = 'subsequence of bytes not found'

    def test_bytes_literal_value(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = b'\x01\x02\x03'
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.STLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytesLiteral.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_bytes_get_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytesGetValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', bytes([1, 2, 3]))
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'Main', b'0')
        self.assertEqual(48, result)

    def test_bytes_get_value_negative_index(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSHM1
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytesGetValueNegativeIndex.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', bytes([1, 2, 3]))
        self.assertEqual(3, result)
        result = self.run_smart_contract(engine, path, 'Main', b'0')
        self.assertEqual(48, result)

    def test_bytes_set_value(self):
        path = self.get_contract_path('BytesSetValue.py')
        self.assertCompilerLogs(CompilerError.UnresolvedOperation, path)

    def test_bytes_clear(self):
        path = self.get_contract_path('BytesClear.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_bytes_reverse(self):
        path = self.get_contract_path('BytesReverse.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_bytes_to_int(self):
        path = self.get_contract_path('BytesToInt.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_int')
        self.assertEqual(513, result)

    def test_bytes_to_int_with_builtin(self):
        path = self.get_contract_path('BytesToIntWithBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_int')
        self.assertEqual(513, result)

    def test_bytes_to_int_mismatched_types(self):
        path = self.get_contract_path('BytesToIntWithBuiltinMismatchedTypes.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_bytes_to_int_with_byte_array_builtin(self):
        path = self.get_contract_path('BytesToIntWithBytearrayBuiltin.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_bytes_to_bool(self):
        path = self.get_contract_path('BytesToBool.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_bool', b'\x00')
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'bytes_to_bool', b'\x01')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'bytes_to_bool', b'\x02')
        self.assertEqual(True, result)

    def test_bytes_to_bool_with_builtin(self):
        path = self.get_contract_path('BytesToBoolWithBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_bool', b'\x00')
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'bytes_to_bool', b'\x01')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'bytes_to_bool', b'\x02')
        self.assertEqual(True, result)

    def test_bytes_to_bool_with_builtin_hard_coded_false(self):
        path = self.get_contract_path('BytesToBoolWithBuiltinHardCodedFalse.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_bool')
        self.assertEqual(False, result)

    def test_bytes_to_bool_with_builtin_hard_coded_true(self):
        path = self.get_contract_path('BytesToBoolWithBuiltinHardCodedTrue.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_bool')
        self.assertEqual(True, result)

    def test_bytes_to_bool_mismatched_types(self):
        path = self.get_contract_path('BytesToBoolWithBuiltinMismatchedTypes.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_bytes_to_str(self):
        path = self.get_contract_path('BytesToStr.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_str')
        self.assertEqual('abc', result)

    def test_bytes_to_str_with_builtin(self):
        path = self.get_contract_path('BytesToStrWithBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_str')
        self.assertEqual('123', result)

    def test_bytes_to_str_mismatched_types(self):
        path = self.get_contract_path('BytesToStrWithBuiltinMismatchedTypes.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_bytes_from_byte_array(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x02'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT + StackItemType.Buffer
            + Opcode.STLOC0
            + Opcode.LDLOC0     # b = a
            + Opcode.STLOC1
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytesFromBytearray.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assign_with_slice(self):
        path = self.get_contract_path('AssignSlice.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', b'unittest',
                                         expected_result_type=bytearray)
        self.assertEqual(b'unittest'[1:2], result)

        result = self.run_smart_contract(engine, path, 'main', b'123',
                                         expected_result_type=bytearray)
        self.assertEqual(b'123'[1:2], result)

        result = self.run_smart_contract(engine, path, 'main', bytearray(),
                                         expected_result_type=bytearray)
        self.assertEqual(bytearray()[1:2], result)

    def test_slice_with_cast(self):
        path = self.get_contract_path('SliceWithCast.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', b'unittest',
                                         expected_result_type=bytes)
        self.assertEqual(b'unittest'[1:2], result)

        result = self.run_smart_contract(engine, path, 'main', '123',
                                         expected_result_type=bytes)
        self.assertEqual(b'123'[1:2], result)

        result = self.run_smart_contract(engine, path, 'main', bytearray(),
                                         expected_result_type=bytearray)
        self.assertEqual(bytearray()[1:2], result)

        result = self.run_smart_contract(engine, path, 'main', 12345,
                                         expected_result_type=bytes)
        self.assertEqual(Integer(12345).to_byte_array()[1:2], result)

    def test_slice_with_stride(self):
        path = self.get_contract_path('SliceWithStride.py')
        engine = TestEngine()

        a = b'unit_test'
        expected_result = a[2:5:2]
        result = self.run_smart_contract(engine, path, 'literal_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-6:5:2]
        result = self.run_smart_contract(engine, path, 'negative_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[0:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-6:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-999:5:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[0:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-999:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[999:5:2]
        result = self.run_smart_contract(engine, path, 'really_high_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[0:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[999:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_slice_with_negative_stride(self):
        path = self.get_contract_path('SliceWithNegativeStride.py')
        engine = TestEngine()

        a = b'unit_test'
        expected_result = a[2:5:-1]
        result = self.run_smart_contract(engine, path, 'literal_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-6:5:-1]
        result = self.run_smart_contract(engine, path, 'negative_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[0:-1:-1]
        result = self.run_smart_contract(engine, path, 'negative_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-6:-1:-1]
        result = self.run_smart_contract(engine, path, 'negative_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-999:5:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[0:-999:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-999:-999:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[999:5:-1]
        result = self.run_smart_contract(engine, path, 'really_high_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[0:999:-1]
        result = self.run_smart_contract(engine, path, 'really_high_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[999:999:-1]
        result = self.run_smart_contract(engine, path, 'really_high_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_slice_omitted_with_stride(self):
        path = self.get_contract_path('SliceOmittedWithStride.py')
        engine = TestEngine()

        a = b'unit_test'
        expected_result = a[::2]
        result = self.run_smart_contract(engine, path, 'omitted_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[:5:2]
        result = self.run_smart_contract(engine, path, 'omitted_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[2::2]
        result = self.run_smart_contract(engine, path, 'omitted_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-6::2]
        result = self.run_smart_contract(engine, path, 'negative_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-999::2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[999::2]
        result = self.run_smart_contract(engine, path, 'really_high_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_slice_omitted_with_negative_stride(self):
        path = self.get_contract_path('SliceOmittedWithNegativeStride.py')
        engine = TestEngine()

        a = b'unit_test'
        expected_result = a[::-2]
        result = self.run_smart_contract(engine, path, 'omitted_values',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[:5:-2]
        result = self.run_smart_contract(engine, path, 'omitted_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[2::-2]
        result = self.run_smart_contract(engine, path, 'omitted_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-6::-2]
        result = self.run_smart_contract(engine, path, 'negative_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[:-1:-2]
        result = self.run_smart_contract(engine, path, 'negative_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[-999::-2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[:-999:-2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[999::-2]
        result = self.run_smart_contract(engine, path, 'really_high_start',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        a = b'unit_test'
        expected_result = a[:999:-2]
        result = self.run_smart_contract(engine, path, 'really_high_end',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_byte_array_get_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytearrayGetValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', bytes([1, 2, 3]))
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'Main', b'0')
        self.assertEqual(48, result)

    def test_byte_array_get_value_negative_index(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSHM1
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytearrayGetValueNegativeIndex.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', bytes([1, 2, 3]))
        self.assertEqual(3, result)
        result = self.run_smart_contract(engine, path, 'Main', b'0')
        self.assertEqual(48, result)

    def test_byte_array_set_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.CONVERT + StackItemType.Buffer
            + Opcode.STLOC0
            + Opcode.LDLOC0     # var[0] = 0x01
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.LDLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytearraySetValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', b'123',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x0123', result)
        result = self.run_smart_contract(engine, path, 'Main', b'0',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x01', result)

    def test_byte_array_set_value_negative_index(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.CONVERT + StackItemType.Buffer
            + Opcode.STLOC0
            + Opcode.LDLOC0     # var[-1] = 0x01
            + Opcode.PUSHM1
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.LDLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytearraySetValueNegativeIndex.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', b'123',
                                         expected_result_type=bytes)
        self.assertEqual(b'12\x01', result)
        result = self.run_smart_contract(engine, path, 'Main', b'0',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x01', result)

    def test_byte_array_literal_value(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.STLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytearrayLiteral.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_byte_array_default(self):
        expected_output = (
            Opcode.PUSH0      # bytearray()
            + Opcode.NEWBUFFER
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytearrayDefault.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'create_bytearray',
                                         expected_result_type=bytearray)
        self.assertEqual(bytearray(), result)

    def test_byte_array_from_literal_bytes(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT + StackItemType.Buffer
            + Opcode.STLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytearrayFromLiteralBytes.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_from_variable_bytes(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x02'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = b'\x01\x02\x03'
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.STLOC0
            + Opcode.PUSHDATA1  # b = bytearray(a)
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT + StackItemType.Buffer
            + Opcode.STLOC1
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytearrayFromVariableBytes.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_from_size(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # bytearray(size)
            + Opcode.NEWBUFFER
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BytearrayFromSize.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'create_bytearray', 10,
                                         expected_result_type=bytearray)
        self.assertEqual(bytearray(10), result)

        result = self.run_smart_contract(engine, path, 'create_bytearray', 0,
                                         expected_result_type=bytearray)
        self.assertEqual(bytearray(0), result)

        # cannot build with negative size
        with self.assertRaisesRegex(TestExecutionException, f'^{self.MAX_ITEM_SIZE_EXCEED_MSG_PREFIX}'):
            result = self.run_smart_contract(engine, path, 'create_bytearray', -10,
                                             expected_result_type=bytes)

    def test_byte_array_from_list_of_int(self):
        path = self.get_contract_path('BytearrayFromListOfInt.py')
        compiler_error_message = self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

        from boa3.model.builtin.builtin import Builtin
        from boa3.model.type.type import Type
        arg_type = Type.list.build([Type.int])
        expected_error = CompilerError.NotSupportedOperation(0, 0, f'{Builtin.ByteArray.identifier}({arg_type.identifier})')
        self.assertEqual(expected_error._error_message, compiler_error_message)

    def test_byte_array_string(self):
        path = self.get_contract_path('BytearrayFromString.py')
        engine = TestEngine()

        # Neo3-boa's bytearray only converts with utf-8 encoding
        string = 'string value'
        result = self.run_smart_contract(engine, path, 'main', string, expected_result_type=bytearray)
        self.assertEqual(bytearray(string, 'utf-8'), result)

        string = 'áãõ😀'
        result = self.run_smart_contract(engine, path, 'main', string, expected_result_type=bytearray)
        self.assertEqual(bytearray(string, 'utf-8'), result)

    def test_byte_array_string_with_encoding(self):
        path = self.get_contract_path('BytearrayFromStringWithEncoding.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_byte_array_append(self):
        path = self.get_contract_path('BytearrayAppend.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x01\x02\x03\x04', result)

    def test_byte_array_append_with_builtin(self):
        path = self.get_contract_path('BytearrayAppendWithBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x01\x02\x03\x04', result)

    def test_byte_array_append_mutable_sequence_with_builtin(self):
        path = self.get_contract_path('BytearrayAppendWithMutableSequence.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x01\x02\x03\x04', result)

    def test_byte_array_clear(self):
        path = self.get_contract_path('BytearrayClear.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'', result)

    def test_byte_array_reverse(self):
        path = self.get_contract_path('BytearrayReverse.py')
        Boa3.compile(path)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x03\x02\x01', result)

    def test_byte_array_extend(self):
        path = self.get_contract_path('BytearrayExtend.py')
        Boa3.compile(path)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x01\x02\x03\x04\x05\x06', result)

    def test_byte_array_extend_with_builtin(self):
        path = self.get_contract_path('BytearrayExtendWithBuiltin.py')
        Boa3.compile(path)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x01\x02\x03\x04\x05\x06', result)

    def test_byte_array_to_int(self):
        path = self.get_contract_path('BytearrayToInt.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_int')
        self.assertEqual(513, result)

    def test_byte_array_to_int_with_builtin(self):
        path = self.get_contract_path('BytearrayToIntWithBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_int')
        self.assertEqual(513, result)

    def test_byte_array_to_int_with_bytes_builtin(self):
        path = self.get_contract_path('BytearrayToIntWithBytesBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_to_int')
        self.assertEqual(513, result)

    def test_boa2_byte_array_test(self):
        path = self.get_contract_path('BytearrayBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\t\x01\x02', result)

    def test_boa2_byte_array_test2(self):
        path = self.get_contract_path('BytearrayBoa2Test2.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_boa2_byte_array_test3(self):
        path = self.get_contract_path('BytearrayBoa2Test3.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(b'\x01\x02\xaa\xfe', result)

    def test_boa2_slice_test(self):
        path = self.get_contract_path('SliceBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x01\x02\x03\x04', result)

    def test_boa2_slice_test2(self):
        path = self.get_contract_path('SliceBoa2Test2.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x02\x03\x04\x02\x03\x04\x05\x06\x01\x02\x03\x04\x03\x04', result)

    def test_uint160_bytes(self):
        path = self.get_contract_path('UInt160Bytes.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        if isinstance(result, str):
            from boa3.neo.vm.type.String import String
            result = String(result).to_bytes()
        self.assertEqual(b'0123456789abcdefghij', result)

    def test_uint160_int(self):
        path = self.get_contract_path('UInt160Int.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        if isinstance(result, str):
            from boa3.neo.vm.type.String import String
            result = String(result).to_bytes()
        self.assertEqual((160).to_bytes(2, 'little') + bytes(18), result)

    def test_uint256_bytes(self):
        path = self.get_contract_path('UInt256Bytes.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        if isinstance(result, str):
            result = String(result).to_bytes()
        self.assertEqual(b'0123456789abcdefghijklmnopqrstuv', result)

    def test_uint256_int(self):
        path = self.get_contract_path('UInt256Int.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        if isinstance(result, str):
            result = String(result).to_bytes()
        self.assertEqual((256).to_bytes(2, 'little') + bytes(30), result)

    def test_bytes_upper(self):
        path = self.get_contract_path('UpperBytesMethod.py')
        engine = TestEngine()

        bytes_value = b'abcdefghijklmnopqrstuvwxyz'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, expected_result_type=bytes)
        self.assertEqual(bytes_value.upper(), result)

        bytes_value = b'a1b123y3z'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, expected_result_type=bytes)
        self.assertEqual(bytes_value.upper(), result)

        bytes_value = b'!@#$%123*-/'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, expected_result_type=bytes)
        self.assertEqual(bytes_value.upper(), result)

    def test_bytes_lower(self):
        path = self.get_contract_path('LowerBytesMethod.py')
        engine = TestEngine()

        bytes_value = b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, expected_result_type=bytes)
        self.assertEqual(bytes_value.lower(), result)

        bytes_value = b'A1B123Y3Z'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, expected_result_type=bytes)
        self.assertEqual(bytes_value.lower(), result)

        bytes_value = b'!@#$%123*-/'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, expected_result_type=bytes)
        self.assertEqual(bytes_value.lower(), result)

    def test_bytes_startswith_method(self):
        path = self.get_contract_path('StartswithBytesMethod.py')
        engine = TestEngine()

        bytes_value = b'unit_test'
        subbytes_value = b'unit'
        start = 0
        end = len(bytes_value)
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start, end)
        self.assertEqual(bytes_value.startswith(subbytes_value, start, end), result)

        bytes_value = b'unit_test'
        subbytes_value = b'unit'
        start = 2
        end = 6
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start, end)
        self.assertEqual(bytes_value.startswith(subbytes_value, start, end), result)

        bytes_value = b'unit_test'
        subbytes_value = b'it'
        start = 2
        end = 6
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start, end)
        self.assertEqual(bytes_value.startswith(subbytes_value, start, end), result)

        bytes_value = b'unit_test'
        subbytes_value = b'it'
        start = 2
        end = 3
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start, end)
        self.assertEqual(bytes_value.startswith(subbytes_value, start, end), result)

        bytes_value = b'unit_test'
        subbytes_value = b'unit_tes'
        start = -99
        end = -1
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start, end)
        self.assertEqual(bytes_value.startswith(subbytes_value, start, end), result)

        bytes_value = b'unit_test'
        subbytes_value = b''
        start = 0
        end = 0
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start, end)
        self.assertEqual(bytes_value.startswith(subbytes_value, start, end), result)

        bytes_value = b'unit_test'
        subbytes_value = b'unit_test'
        start = 0
        end = 99
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start, end)
        self.assertEqual(bytes_value.startswith(subbytes_value, start, end), result)

    def test_bytes_startswith_method_default_end(self):
        path = self.get_contract_path('StartswithBytesMethodDefaultEnd.py')
        engine = TestEngine()

        bytes_value = b'unit_test'
        subbytes_value = b'unit'
        start = 0
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start)
        self.assertEqual(bytes_value.startswith(subbytes_value, start), result)

        bytes_value = b'unit_test'
        subbytes_value = b'unit'
        start = 2
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start)
        self.assertEqual(bytes_value.startswith(subbytes_value, start), result)

        bytes_value = b'unit_test'
        subbytes_value = b'it'
        start = 2
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start)
        self.assertEqual(bytes_value.startswith(subbytes_value, start), result)

        bytes_value = b'unit_test'
        subbytes_value = b'it'
        start = 3
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start)
        self.assertEqual(bytes_value.startswith(subbytes_value, start), result)

        bytes_value = b'unit_test'
        subbytes_value = b'unit_tes'
        start = -99
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start)
        self.assertEqual(bytes_value.startswith(subbytes_value, start), result)

        bytes_value = b'unit_test'
        subbytes_value = b''
        start = 0
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start)
        self.assertEqual(bytes_value.startswith(subbytes_value, start), result)

        bytes_value = b'unit_test'
        subbytes_value = b''
        start = 99
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start)
        self.assertEqual(bytes_value.startswith(subbytes_value, start), result)

        bytes_value = b'unit_test'
        subbytes_value = b'unit_test'
        start = 0
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value, start)
        self.assertEqual(bytes_value.startswith(subbytes_value, start), result)

    def test_bytes_startswith_method_defaults(self):
        path = self.get_contract_path('StartswithBytesMethodDefaults.py')
        engine = TestEngine()

        bytes_value = b'unit_test'
        subbytes_value = b'unit'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value)
        self.assertEqual(bytes_value.startswith(subbytes_value), result)

        bytes_value = b'unit_test'
        subbytes_value = b'unit_test'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value)
        self.assertEqual(bytes_value.startswith(subbytes_value), result)

        bytes_value = b'unit_test'
        subbytes_value = b''
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value)
        self.assertEqual(bytes_value.startswith(subbytes_value), result)

        bytes_value = b'unit_test'
        subbytes_value = b'12345'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value)
        self.assertEqual(bytes_value.startswith(subbytes_value), result)

        bytes_value = b'unit_test'
        subbytes_value = b'bigger subbytes_value'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, subbytes_value)
        self.assertEqual(bytes_value.startswith(subbytes_value), result)

    def test_bytes_strip(self):
        path = self.get_contract_path('StripBytesMethod.py')
        engine = TestEngine()

        bytes_value = b'abcdefghijklmnopqrstuvwxyz'
        sub_bytes = b'abcxyz'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, sub_bytes,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.strip(sub_bytes), result)

        bytes_value = b'abcdefghijklmnopqrsvwxyz unit test abcdefghijklmnopqrsvwxyz'
        sub_bytes = b'abcdefghijklmnopqrsvwxyz '
        result = self.run_smart_contract(engine, path, 'main', bytes_value, sub_bytes,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.strip(sub_bytes), result)

        bytes_value = b'0123456789hello world987654310'
        sub_bytes = b'0987654321'
        result = self.run_smart_contract(engine, path, 'main', bytes_value, sub_bytes,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.strip(sub_bytes), result)

    def test_bytes_strip_default(self):
        path = self.get_contract_path('StripBytesMethodDefault.py')
        engine = TestEngine()

        bytes_value = b'     unit test    '
        result = self.run_smart_contract(engine, path, 'main', bytes_value,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.strip(), result)

        bytes_value = b'unit test    '
        result = self.run_smart_contract(engine, path, 'main', bytes_value,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.strip(), result)

        bytes_value = b'    unit test'
        result = self.run_smart_contract(engine, path, 'main', bytes_value,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.strip(), result)

        bytes_value = b' \t\n\r\f\vunit test \t\n\r\f\v'
        result = self.run_smart_contract(engine, path, 'main', bytes_value,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.strip(), result)

    def test_isdigit_method(self):
        path = self.get_contract_path('IsdigitMethod.py')
        engine = TestEngine()

        bytes_value = b'0123456789'
        result = self.run_smart_contract(engine, path, 'main', bytes_value)
        self.assertEqual(bytes_value.isdigit(), result)

        bytes_value = b'23mixed01'
        result = self.run_smart_contract(engine, path, 'main', bytes_value)
        self.assertEqual(bytes_value.isdigit(), result)

        bytes_value = b'no digits here'
        result = self.run_smart_contract(engine, path, 'main', bytes_value)
        self.assertEqual(bytes_value.isdigit(), result)

        bytes_value = b''
        result = self.run_smart_contract(engine, path, 'main', bytes_value)
        self.assertEqual(bytes_value.isdigit(), result)

    def test_bytes_join_with_sequence(self):
        path = self.get_contract_path('JoinBytesMethodWithSequence.py')
        engine = TestEngine()

        bytes_value = b' '
        sequence = [b"Unit", b"Test", b"Neo3-boa"]
        result = self.run_smart_contract(engine, path, 'main', bytes_value, sequence,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.join(sequence), result)

        bytes_value = b' '
        sequence = []
        result = self.run_smart_contract(engine, path, 'main', bytes_value, sequence,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.join(sequence), result)

        bytes_value = b' '
        sequence = [b"UnitTest"]
        result = self.run_smart_contract(engine, path, 'main', bytes_value, sequence,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.join(sequence), result)

    def test_bytes_join_with_dictionary(self):
        path = self.get_contract_path('JoinBytesMethodWithDictionary.py')
        engine = TestEngine()

        bytes_value = b' '
        dictionary = {b"Unit": 1, b"Test": 2, b"Neo3-boa": 3}
        result = self.run_smart_contract(engine, path, 'main', bytes_value, dictionary,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.join(dictionary), result)

        bytes_value = b' '
        dictionary = {}
        result = self.run_smart_contract(engine, path, 'main', bytes_value, dictionary,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.join(dictionary), result)

        bytes_value = b' '
        dictionary = {b"UnitTest": 1}
        result = self.run_smart_contract(engine, path, 'main', bytes_value, dictionary,
                                         expected_result_type=bytes)
        self.assertEqual(bytes_value.join(dictionary), result)

    def test_bytes_index(self):
        path = self.get_contract_path('IndexBytes.py')
        engine = TestEngine()

        bytes_ = b'unit test'
        subsequence = b'i'
        start = 0
        end = 4
        result = self.run_smart_contract(engine, path, 'main', bytes_, subsequence, start, end)
        self.assertEqual(bytes_.index(subsequence, start, end), result)

        bytes_ = b'unit test'
        bytes_sequence = b'i'
        start = 2
        end = 4
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence, start, end)
        self.assertEqual(bytes_.index(bytes_sequence, start, end), result)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSEQUENCE_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 'i', 3, 4)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSEQUENCE_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 'i', 4, -1)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSEQUENCE_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 'i', 0, -99)

        bytes_ = b'unit test'
        bytes_sequence = b'i'
        start = 0
        end = -1
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence, start, end)
        self.assertEqual(bytes_.index(bytes_sequence, start, end), result)

        bytes_ = b'unit test'
        bytes_sequence = b'n'
        start = 0
        end = 99
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence, start, end)
        self.assertEqual(bytes_.index(bytes_sequence, start, end), result)

    def test_bytes_index_end_default(self):
        path = self.get_contract_path('IndexBytesEndDefault.py')
        engine = TestEngine()

        bytes_ = b'unit test'
        bytes_sequence = b't'
        start = 0
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence, start)
        self.assertEqual(bytes_.index(bytes_sequence, start), result)

        bytes_ = b'unit test'
        bytes_sequence = b't'
        start = 4
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence, start)
        self.assertEqual(bytes_.index(bytes_sequence, start), result)

        bytes_ = b'unit test'
        bytes_sequence = b't'
        start = 6
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence, start)
        self.assertEqual(bytes_.index(bytes_sequence, start), result)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSEQUENCE_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 'i', 99)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSEQUENCE_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 't', -1)

        bytes_ = b'unit test'
        bytes_sequence = b'i'
        start = -10
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence, start)
        self.assertEqual(bytes_.index(bytes_sequence, start), result)

    def test_bytes_index_defaults(self):
        path = self.get_contract_path('IndexBytesDefaults.py')
        engine = TestEngine()

        bytes_ = b'unit test'
        bytes_sequence = b'u'
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence)
        self.assertEqual(bytes_.index(bytes_sequence), result)

        bytes_ = b'unit test'
        bytes_sequence = b't'
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence)
        self.assertEqual(bytes_.index(bytes_sequence), result)

        bytes_ = b'unit test'
        bytes_sequence = b' '
        result = self.run_smart_contract(engine, path, 'main', bytes_, bytes_sequence)
        self.assertEqual(bytes_.index(bytes_sequence), result)

    def test_bytes_index_mismatched_type(self):
        path = self.get_contract_path('IndexBytesMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)



from __future__ import annotations

from typing import Any, Dict

from boa3.model.builtin.interop.nativecontract import ContractManagement
from boa3.model.builtin.native.inativecontractclass import INativeContractClass
from boa3.model.method import Method


class ContractManagementClass(INativeContractClass):
    """
    A class used to represent ContractManagement native contract
    """

    def __init__(self):
        super().__init__('ContractManagement', ContractManagement)

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.model.builtin.interop.interop import Interop

        if len(self._class_methods) == 0:
            from boa3.model.builtin.native.contract_management import HasMethod

            self._class_methods = {
                'get_minimum_deployment_fee': Interop.GetMinimumDeploymentFee,
                'get_contract': Interop.GetContract,
                'has_method': HasMethod(),
                'deploy': Interop.CreateContract,
                'update': Interop.UpdateContract,
                'destroy': Interop.DestroyContract
            }
        return super().class_methods

    @classmethod
    def build(cls, value: Any = None) -> ContractManagementClass:
        if value is None or cls._is_type_of(value):
            return _ContractManagement

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, ContractManagementClass)


_ContractManagement = ContractManagementClass()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Optional

from boa3.model.operation.operation import IOperation
from boa3.model.type.itype import IType


class UnaryOperation(IOperation, ABC):
    """
    An interface used to represent binary operations

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar operand: the operand type.
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = []

    def __init__(self, operand: IType):
        self.operand_type: IType = operand
        result = self._get_result(operand)
        super().__init__(self.operator, result)

    @property
    def number_of_operands(self) -> int:
        return 1

    @abstractmethod
    def _get_result(self, operand: IType) -> IType:
        """
        Gets the result type of the operation given the operand type.

        :param operand:  operand type
        :return: the result type of the operation. Type.none if the operand is not valid.
        """
        pass

    @classmethod
    def build(cls, operand: IType) -> Optional[UnaryOperation]:
        """
        Creates a unary operation with the given operand type

        :param operand: operand type
        :return: The built operation if the operands are valid. None otherwise
        :rtype: UnaryOperation or None
        """
        return cls(operand)



from test_runner.neoxp.command.neoexpresscommand import NeoExpressCommand


def create_checkpoint(checkpoint_path: str, force: bool = False) -> NeoExpressCommand:
    options = {}
    if force:
        options['--force'] = ''

    return NeoExpressCommand('checkpoint create', [checkpoint_path], options)


def fastfwd(block_count: int) -> NeoExpressCommand:
    if not isinstance(block_count, int) or block_count < 1:
        block_count = 1

    options = {}
    return NeoExpressCommand('fastfwd', [str(block_count)], options)



class Example:
    def __init__(self, id_: int, arg1: str):
        self._var_int = id_
        self.var_str = arg1

    def export(self) -> dict:
        return {
            'text': self.var_str
        }



from __future__ import annotations

from typing import Any, Dict

from boa3.constants import GAS_SCRIPT
from boa3.model.builtin.interop.contract import GasToken
from boa3.model.builtin.native.inativecontractclass import INativeContractClass
from boa3.model.method import Method


class GasClass(INativeContractClass):
    """
    A class used to represent GAS native contract
    """

    def __init__(self):
        super().__init__('GAS', GasToken)

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.model.builtin.native.nep17_methods import (BalanceOfMethod, DecimalsMethod, SymbolMethod,
                                                             TotalSupplyMethod, TransferMethod)

        if len(self._class_methods) == 0:
            self._class_methods = {
                'balanceOf': BalanceOfMethod(GAS_SCRIPT),
                'decimals': DecimalsMethod(GAS_SCRIPT),
                'symbol': SymbolMethod(GAS_SCRIPT),
                'totalSupply': TotalSupplyMethod(GAS_SCRIPT),
                'transfer': TransferMethod(GAS_SCRIPT)
            }
        return super().class_methods

    @classmethod
    def build(cls, value: Any = None) -> GasClass:
        if value is None or cls._is_type_of(value):
            return _Gas

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, GasClass)


_Gas = GasClass()



import ast
from typing import Dict, List, Tuple, Optional

from boa3.compiler.compiledmetadata import CompiledMetadata
from boa3.model.builtin.interop.contractgethashmethod import ContractGetHashMethod
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.opcode.OpcodeInfo import OpcodeInfo


class NativeContractMethod(InteropMethod):

    def __init__(self, native_contract_script_hash_method: ContractGetHashMethod, identifier: str, syscall: str,
                 args: Dict[str, Variable] = None, defaults: List[ast.AST] = None, return_type: IType = None,
                 internal_call_args: int = None):
        super().__init__(identifier, syscall, args, defaults, return_type)
        self.script_hash_method = native_contract_script_hash_method
        min_no_args = len(self.args_without_default)
        if not isinstance(internal_call_args, int):
            internal_call_args = len(self.args)
        elif len(self.args_without_default) > internal_call_args:
            internal_call_args = min_no_args

        self.internal_call_args = internal_call_args

        from boa3.neo3.contracts.contracttypes import CallFlags
        self._call_flags_default = CallFlags.ALL
        self._pack_arguments = None   # defined during compilation
        self._method_token_id = None  # defined during compilation
        self.external_name = self._sys_call

    @property
    def contract_script_hash(self) -> bytes:
        return self.script_hash_method.script_hash

    def reset(self):
        # reset the object state to ensure the correct output when calling consecutive compilations
        super().reset()
        self._pack_arguments = None
        self._method_token_id = None

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.builtin.interop.interop import Interop
        from boa3.model.type.type import Type
        from boa3.neo.vm.type.Integer import Integer
        from boa3.neo.vm.type.String import String

        call_flag = self._call_flags_default
        if self._pack_arguments is None:
            self._pack_arguments = False

        CompiledMetadata.instance().add_contract_permission(self.contract_script_hash, self._sys_call)
        self_method_token_id = self._get_method_token_id(call_flag)

        if isinstance(self_method_token_id, int) and self_method_token_id >= 0:
            call_opcode = Opcode.CALLT
            opcode_info = OpcodeInfo.get_info(call_opcode)
            arg_size = opcode_info.data_len

            call_opcodes = [
                (call_opcode, Integer(self_method_token_id).to_byte_array(min_length=arg_size))
            ]
        else:
            if len(self.args) == 0:
                pack_args = [(Opcode.NEWARRAY0, b'')]
            else:
                from boa3.neo.vm.opcode import OpcodeHelper
                pack_args = [
                    OpcodeHelper.get_push_and_data(len(self.args)),
                    (Opcode.PACK, b'')
                ]

            call_flags = Integer(call_flag).to_byte_array(signed=True, min_length=1)
            flags_opcode = [
                (Opcode.PUSHDATA1, Integer(len(call_flags)).to_byte_array() + call_flags)
            ]

            method = String(self._sys_call).to_bytes()
            method_opcode = [
                (Opcode.PUSHDATA1, Integer(len(method)).to_byte_array(min_length=1) + method)
            ]

            drop_if_void_opcode = [
                (Opcode.DROP, b'')
            ] if self.return_type is Type.none else []

            call_opcodes = (pack_args
                            + flags_opcode
                            + method_opcode
                            + self.script_hash_method.opcode
                            + Interop.CallContract.opcode
                            + drop_if_void_opcode
                            )

        return call_opcodes

    def _get_method_token_id(self, call_flag=None) -> Optional[int]:
        if self._method_token_id is None:
            from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping

            if call_flag is None:
                call_flag = self._call_flags_default
            self._method_token_id = VMCodeMapping.instance().add_method_token(self, call_flag)
        return self._method_token_id

    @property
    def pack_arguments(self) -> bool:
        if self._pack_arguments is None:
            from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
            self_method_token_id = self._get_method_token_id(self._call_flags_default)

            self._pack_arguments = not isinstance(self_method_token_id, int) or self_method_token_id < 0
        return self._pack_arguments

    @property
    def method_name(self) -> str:
        return self._sys_call



from typing import List, Optional

from boa3.builtin.type import ECPoint, UInt160
from boa3.neo.vm.type.ContractParameterType import ContractParameterType


class ContractManifest:
    """
    Represents the manifest of a smart contract [#]_.

    When a smart contract is deployed, it must explicitly declare the features and permissions it will use.

    When it is running, it will be limited by its declared list of features and permissions, and cannot make any
    behavior beyond the scope of the list.

    .. [#] For more details, see NEP-15.

    :ivar name: The name of the contract.
    :vartype name: str
    :ivar groups: The groups of the contract.
    :vartype groups: List[ContractGroup]
    :ivar supported_standards: Indicates which standards the contract supports. It can be a list of NEPs.
    :vartype supported_standards: List[str]
    :ivar abi: The ABI of the contract.
    :vartype abi: ContractAbi
    :ivar permissions: The permissions of the contract.
    :vartype permissions: List[ContractPermission]
    :ivar trusts:
        The trusted contracts and groups of the contract.

        If a contract is trusted, the user interface will not give any warnings when called by the contract.

    :vartype trusts: List[ContractPermissionDescriptor] or None
    :ivar extras: Custom user data as a json string.
    :vartype extras: str
    """

    def __init__(self):
        self.name: str = ''
        self.groups: List[ContractGroup] = []
        self.supported_standards: List[str] = []
        self.abi: ContractAbi = None
        self.permissions: List[ContractPermission] = []
        self.trusts: Optional[List[ContractPermissionDescriptor]] = None
        self.extras: str = ''


class ContractPermission:
    """
    Represents a permission of a contract. It describes which contracts may be invoked and which methods are called.

    If a contract invokes a contract or method that is not declared in the manifest at runtime, the invocation will
    fail.

    :ivar contract:
        Indicates which contract to be invoked.

        It can be a hash of a contract, a public key of a group, or a wildcard *.

        If it specifies a hash of a contract, then the contract will be invoked; If it specifies a public key of a
        group, then any contract in this group may be invoked; If it specifies a wildcard *, then any contract may be
        invoked.

    :vartype contract: ContractPermissionDescriptor or None
    :ivar methods:
        Indicates which methods to be called.

        It can also be assigned with a wildcard *. If it is a wildcard *, then it means that any method can be called.

    :vartype methods: List[str] or None
    """

    def __init__(self):
        self.contract: Optional[ContractPermissionDescriptor] = None
        self.methods: Optional[List[str]] = None


class ContractPermissionDescriptor:
    """
    Indicates which contracts are authorized to be called.

    :ivar hash: The hash of the contract.
    :vartype hash: UInt160 or None
    :ivar group: The group of the contracts.
    :vartype group: ECPoint or None
    """

    def __init__(self):
        self.hash: Optional[UInt160] = None
        self.group: Optional[ECPoint] = None


class ContractGroup:
    """
    Represents a set of mutually trusted contracts.

    A contract will trust and allow any contract in the same group to invoke it, and the user interface will not give
    any warnings.

    A group is identified by a public key and must be accompanied by a signature for the contract hash to prove that
    the contract is indeed included in the group.

    :ivar pubkey: The public key of the group.
    :vartype pubkey: ECPoint
    :ivar signature: The signature of the contract hash which can be verified by `pubkey`.
    :vartype signature: bytes
    """

    def __init__(self):
        self.pubkey: ECPoint = ECPoint(b'')
        self.signature: bytes = b''


class ContractAbi:
    """
    Represents the ABI of a smart contract [#]_.

    .. [#] For more details, see NEP-14.

    :ivar methods: Gets the methods in the ABI.
    :vartype methods: List[ContractMethodDescriptor]
    :ivar events: Gets the events in the ABI.
    :vartype events: List[ContractEventDescriptor]
    """

    def __init__(self):
        self.methods: List[ContractMethodDescriptor] = []
        self.events: List[ContractEventDescriptor] = []


class ContractMethodDescriptor:
    """
    Represents a method in a smart contract ABI.

    :ivar name: The name of the method.
    :vartype name: str
    :ivar parameters: The parameters of the method.
    :vartype parameters: List[ContractParameterDefinition]
    :ivar return_type: Indicates the return type of the method.
    :vartype return_type: ContractParameterType
    :ivar offset: The position of the method in the contract script.
    :vartype offset: int
    :ivar safe:
        Indicates whether the method is a safe method.

        If a method is marked as safe, the user interface will not give any warnings when it is called by other
        contracts.

    :vartype safe: bool
    """

    def __init__(self):
        self.name: str = ''
        self.parameters: List[ContractParameterDefinition] = []
        self.return_type: ContractParameterType = ContractParameterType.Any
        self.offset: int = 0
        self.safe: bool = False


class ContractEventDescriptor:
    """
    Represents an event in a smart contract ABI.

    :ivar name: The name of the event.
    :vartype name: str
    :ivar parameters: The parameters of the event.
    :vartype parameters: List[ContractParameterDefinition]
    """

    def __init__(self):
        self.name: str = ''
        self.parameters: List[ContractParameterDefinition] = []


class ContractParameterDefinition:
    """
    Represents a parameter of an event or method in ABI.

    :ivar name: The name of the parameter.
    :vartype name: str
    :ivar type: The type of the parameter.
    :vartype type: ContractParameterType
    """

    def __init__(self):
        self.name: str = ''
        self.type: ContractParameterType = ContractParameterType.Any



__all__ = ['DecimalCeilingMethod',
           'DecimalFloorMethod',
           'PowMethod',
           'SqrtMethod',
           ]

from boa3.model.builtin.math.decimalceil import DecimalCeilingMethod
from boa3.model.builtin.math.decimalfloor import DecimalFloorMethod
from boa3.model.builtin.math.powmethod import PowMethod
from boa3.model.builtin.math.sqrtmethod import SqrtMethod



import json
from typing import Dict, List

from boa3.neo import to_script_hash
from boa3.neo.vm.type.String import String
from boa3.neo3.core.types import UInt160
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestNEP11Template(BoaTest):
    default_folder: str = 'examples'

    OWNER_SCRIPT_HASH = to_script_hash(b'NZcuGiwRu1QscpmCyxj5XwQBUf6sk7dJJN')
    OTHER_ACCOUNT_1 = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')
    OTHER_ACCOUNT_2 = to_script_hash(b'NNLi44dJNXtDNSBkofB48aTVYtb1zZrNEs')

    TOKEN_META = bytes(
        '{ "name": "NEP11", "description": "Some description", "image": "{some image URI}", "tokenURI": "{some URI}" }',
        'utf-8')
    TOKEN_LOCKED = bytes('lockedContent', 'utf-8')
    ROYALTIES = bytes(
        '[{"address": "NZcuGiwRu1QscpmCyxj5XwQBUf6sk7dJJN", "value": 2000}, '
        '{"address": "NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB", "value": 3000}]',
        'utf-8')

    def deploy_contract(self, engine, path):
        engine.add_signer_account(self.OWNER_SCRIPT_HASH)
        self.run_smart_contract(engine, path, '_deploy', self.OWNER_SCRIPT_HASH, False,
                                signer_accounts=[self.OWNER_SCRIPT_HASH],
                                expected_result_type=bool)

    def test_nep11_compile(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('supportedstandards', manifest)
        self.assertIsInstance(manifest['supportedstandards'], list)
        self.assertGreater(len(manifest['supportedstandards']), 0)
        self.assertIn('NEP-11', manifest['supportedstandards'])

    def test_nep11_symbol(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)
        result = self.run_smart_contract(engine, path, 'symbol')
        self.assertEqual('EXMP', result)

    def test_nep11_decimals(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)
        result = self.run_smart_contract(engine, path, 'decimals')
        self.assertEqual(0, result)

    def test_nep11_total_supply(self):
        # smart contract deploys with zero tokens minted
        total_supply = 0

        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)
        result = self.run_smart_contract(engine, path, 'totalSupply')
        self.assertEqual(total_supply, result)

    def test_nep11_deploy(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)
        # contract is already deployed

        result = self.run_smart_contract(engine, path, '_deploy', None, False,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH],
                                         expected_result_type=bool)
        self.assertIsVoid(result)
        # contract was deployed only once
        self.assertEqual(1, len([notification for notification in engine.notifications if notification.name == "Deploy"]))

    def test_nep11_update(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        new_nef, new_manifest = self.get_bytes_output(path)
        arg_manifest = String(json.dumps(new_manifest, separators=(',', ':'))).to_bytes()

        # update contract
        result = self.run_smart_contract(engine, path, 'update',
                                         new_nef, arg_manifest,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertIsVoid(result)

    def test_nep11_destroy(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        # destroy contract
        result = self.run_smart_contract(engine, path, 'destroy',
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertIsVoid(result)

        # should not exist anymore
        result = self.run_smart_contract(engine, path, 'symbol')
        self.assertNotEqual('', result)

    def test_nep11_verify(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        # should fail because account does not have enough for fees
        verified = self.run_smart_contract(engine, path, 'verify',
                                           signer_accounts=[self.OTHER_ACCOUNT_2],
                                           expected_result_type=bool)
        self.assertEqual(verified, False)

        addresses = self.run_smart_contract(engine, path, 'getAuthorizedAddress',
                                            expected_result_type=List[UInt160],
                                            signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(addresses[0], self.OWNER_SCRIPT_HASH)
        self.assertEqual(len(addresses), 1)

        verified = self.run_smart_contract(engine, path, 'verify',
                                           signer_accounts=[self.OTHER_ACCOUNT_1],
                                           expected_result_type=bool)
        self.assertEqual(verified, False)

        verified = self.run_smart_contract(engine, path, 'verify',
                                           signer_accounts=[self.OTHER_ACCOUNT_2],
                                           expected_result_type=bool)
        self.assertEqual(verified, False)

        verified = self.run_smart_contract(engine, path, 'verify',
                                           signer_accounts=[self.OWNER_SCRIPT_HASH],
                                           expected_result_type=bool)
        self.assertEqual(verified, True)

    def test_nep11_authorize_2(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        self.run_smart_contract(engine, path, 'setAuthorizedAddress',
                                self.OTHER_ACCOUNT_1, True,
                                signer_accounts=[self.OWNER_SCRIPT_HASH])
        auth_events = engine.get_events('Authorized')

        # check if the event was triggered and the address was authorized
        self.assertEqual(0, auth_events[0].arguments[1])
        self.assertEqual(1, auth_events[0].arguments[2])

        # now deauthorize the address
        self.run_smart_contract(engine, path, 'setAuthorizedAddress',
                                self.OTHER_ACCOUNT_1, False,
                                signer_accounts=[self.OWNER_SCRIPT_HASH])
        auth_events = engine.get_events('Authorized')
        # check if the event was triggered and the address was authorized
        self.assertEqual(0, auth_events[1].arguments[1])
        self.assertEqual(0, auth_events[1].arguments[2])

    def test_nep11_authorize(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        self.run_smart_contract(engine, path, 'setAuthorizedAddress',
                                self.OTHER_ACCOUNT_1, True,
                                signer_accounts=[self.OWNER_SCRIPT_HASH])
        auth_events = engine.get_events('Authorized')

        # check if the event was triggered and the address was authorized
        self.assertEqual(0, auth_events[0].arguments[1])
        self.assertEqual(1, auth_events[0].arguments[2])

        # now deauthorize the address
        self.run_smart_contract(engine, path, 'setAuthorizedAddress',
                                self.OTHER_ACCOUNT_1, False,
                                signer_accounts=[self.OWNER_SCRIPT_HASH])
        auth_events = engine.get_events('Authorized')
        # check if the event was triggered and the address was authorized
        self.assertEqual(0, auth_events[1].arguments[1])
        self.assertEqual(0, auth_events[1].arguments[2])

    def test_nep11_pause(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        engine.add_contract(path.replace('.py', '.nef'))

        # add some gas for fees
        add_amount = 10 * 10 ** 8
        engine.add_gas(self.OTHER_ACCOUNT_1, add_amount)

        # pause contract
        result = self.run_smart_contract(engine, path, 'updatePause', True,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH],
                                         expected_result_type=bool)
        self.assertEqual(True, result)

        # should fail because contract is paused
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'mint',
                                    self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                    signer_accounts=[self.OTHER_ACCOUNT_1])

        # unpause contract
        result = self.run_smart_contract(engine, path, 'updatePause', False,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH],
                                         expected_result_type=bool)
        self.assertEqual(False, result)

        # mint
        result = self.run_smart_contract(engine, path, 'mint',
                                         self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual('\x01', result)

    def test_nep11_mint(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        engine.add_contract(path.replace('.py', '.nef'))

        # add some gas for fees
        add_amount = 10 * 10 ** 8
        engine.add_gas(self.OTHER_ACCOUNT_1, add_amount)

        # should succeed now that account has enough fees
        token = self.run_smart_contract(engine, path, 'mint',
                                        self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                        signer_accounts=[self.OTHER_ACCOUNT_1])

        properties = self.run_smart_contract(engine, path, 'properties', token, expected_result_type=Dict[str, str])
        token_property = json.loads(self.TOKEN_META.decode('utf-8').replace("'", "\""))
        self.assertEqual(token_property, properties)

        royalties = self.run_smart_contract(engine, path, 'getRoyalties', token, expected_result_type=str)
        self.assertEqual(royalties, self.ROYALTIES.decode('utf-8'))

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'properties',
                                    bytes('thisisanonexistingtoken', 'utf-8'),
                                    expected_result_type=str)

        # check balances after
        nep11_balance_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        nep11_supply_after = self.run_smart_contract(engine, path, 'totalSupply')
        self.assertEqual(1, nep11_balance_after)
        self.assertEqual(1, nep11_supply_after)

    def test_nep11_transfer(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        engine.add_contract(path.replace('.py', '.nef'))

        # add some gas for fees
        add_amount = 10 * 10 ** 8
        engine.add_gas(self.OTHER_ACCOUNT_1, add_amount)

        # mint
        token = self.run_smart_contract(engine, path, 'mint',
                                        self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                        signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual('\x01', token)

        # check owner before
        nep11_owner_of_before = self.run_smart_contract(engine, path, 'ownerOf', token)
        self.assertEqual(self.OTHER_ACCOUNT_1, nep11_owner_of_before)

        # transfer
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OTHER_ACCOUNT_2, token, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1],
                                         expected_result_type=bool)
        self.assertEqual(True, result)

        # check owner after
        nep11_owner_of_after = self.run_smart_contract(engine, path, 'ownerOf', token)
        self.assertEqual(self.OTHER_ACCOUNT_2, nep11_owner_of_after)

        # check balances after
        nep11_balance_after_transfer = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        nep11_supply_after_transfer = self.run_smart_contract(engine, path, 'totalSupply')
        self.assertEqual(0, nep11_balance_after_transfer)
        self.assertEqual(1, nep11_supply_after_transfer)

        # try to transfer non existing token id
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    self.OTHER_ACCOUNT_2, bytes('thisisanonexistingtoken', 'utf-8'), None,
                                    signer_accounts=[self.OTHER_ACCOUNT_1])

    def test_nep11_burn(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        engine.add_contract(path.replace('.py', '.nef'))

        # add some gas for fees
        add_amount = 10 * 10 ** 8
        engine.add_gas(self.OTHER_ACCOUNT_1, add_amount)

        # mint
        token = self.run_smart_contract(engine, path, 'mint',
                                        self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                        signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual('\x01', token)

        # burn
        burn = self.run_smart_contract(engine, path, 'burn', token,
                                       signer_accounts=[self.OTHER_ACCOUNT_1],
                                       expected_result_type=bool)
        self.assertEqual(True, burn)

        # check balances after
        nep11_balance_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(0, nep11_balance_after)
        nep11_supply_after = self.run_smart_contract(engine, path, 'totalSupply')
        self.assertEqual(0, nep11_supply_after)

    def test_nep11_onNEP11Payment(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)
        engine.add_contract(path.replace('.py', '.nef'))

        # add some gas for fees
        add_amount = 10 * 10 ** 8
        engine.add_gas(self.OTHER_ACCOUNT_1, add_amount)

        # mint
        token = self.run_smart_contract(engine, path, 'mint',
                                        self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                        signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual('\x01', token)

        # the smart contract will abort if any address calls the NEP11 onPayment method
        with self.assertRaisesRegex(TestExecutionException, self.ABORTED_CONTRACT_MSG):
            self.run_smart_contract(engine, path, 'onNEP11Payment',
                                    self.OTHER_ACCOUNT_1, 1, token, None,
                                    signer_accounts=[self.OTHER_ACCOUNT_1])

    def test_nep11_balance_of(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        # add some gas for fees
        add_amount = 10 * 10 ** 8
        engine.add_gas(self.OTHER_ACCOUNT_1, add_amount)

        # mint
        token = self.run_smart_contract(engine, path, 'mint',
                                        self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                        signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual('\x01', token)

        # balance should be one
        amount_of_tokens = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(1, amount_of_tokens)

        # minting another token
        token = self.run_smart_contract(engine, path, 'mint',
                                        self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                        signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual('\x02', token)

        # balance should increase again
        amount_of_tokens = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(2, amount_of_tokens)

    def test_nep11_tokens_of(self):
        path = self.get_contract_path('nep11_non_divisible.py')
        engine = TestEngine()
        self.deploy_contract(engine, path)

        # add some gas for fees
        add_amount = 10 * 10 ** 8
        engine.add_gas(self.OTHER_ACCOUNT_1, add_amount)

        tokens = []

        # initial tokensOf is an empty iterator
        account_tokens = self.run_smart_contract(engine, path, 'tokensOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(tokens, account_tokens)

        # mint
        token = self.run_smart_contract(engine, path, 'mint',
                                        self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                        signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual('\x01', token)

        tokens.append('\x01')

        # tokensOf should return ['\x01']
        account_tokens = self.run_smart_contract(engine, path, 'tokensOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(tokens, account_tokens)

        # minting another token
        token = self.run_smart_contract(engine, path, 'mint',
                                        self.OTHER_ACCOUNT_1, self.TOKEN_META, self.TOKEN_LOCKED, self.ROYALTIES,
                                        signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual('\x02', token)

        tokens.append('\x02')

        # tokens should increase again
        account_tokens = self.run_smart_contract(engine, path, 'tokensOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(tokens, account_tokens)



from abc import ABC
from typing import Any, Set

from boa3.model.type.collection.icollection import ICollectionType
from boa3.model.type.itype import IType
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.StackItem import StackItemType


class SequenceType(ICollectionType, ABC):
    """
    An interface used to represent Python sequence type
    """

    def __init__(self, identifier: str, values_type: Set[IType]):
        super().__init__(identifier, values_type=values_type)
        self.value_type: IType = self.item_type

    @property
    def default_value(self) -> Any:
        return []

    def is_type_of(self, value: Any) -> bool:
        if self._is_type_of(value):
            if isinstance(value, SequenceType):
                return self.value_type.is_type_of(value.value_type)
            return True
        return False

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Array

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.Array



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Dict, List, Tuple

from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.symbol import ISymbol
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.StackItem import StackItemType


class IType(IdentifiedSymbol):
    """
    An interface used to represent types

    :ivar identifier: the name identifier of the type
    """

    def __init__(self, identifier: str):
        super().__init__(identifier)

    @property
    def shadowing_name(self) -> str:
        return 'type'

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self.identifier

    @property
    def default_value(self) -> Any:
        return None

    @property
    def abi_type(self) -> AbiType:
        """
        Get the type representation for the abi

        :return: the representation for the abi. Any by default.
        """
        return AbiType.Any

    @property
    def stack_item(self) -> StackItemType:
        """
        Get the Neo VM stack item type representation for this type

        :return: the stack item type of this type. Any by default.
        """
        return StackItemType.Any

    @property
    def is_generic(self) -> bool:
        """
        Verifies if this is a generic type

        :return: True if this is a generic type. False otherwise.
        """
        return False

    @classmethod
    @abstractmethod
    def _is_type_of(cls, value: Any) -> bool:
        """
        Validates if this is the type of the given value

        :param value: value to check the type
        :return: a boolean value that represents if this is the type of the value.
        :rtype: bool
        """
        pass

    def is_type_of(self, value: Any) -> bool:
        """
        Validates if this is the type of the given value. If this is not a primitive type, validates if its attributes
        are also valid.

        :param value: value to check the type
        :return: a boolean value that represents if this is the type of the value.
        :rtype: bool
        """
        return self._is_type_of(value)

    @classmethod
    @abstractmethod
    def build(cls, value: Any) -> IType:
        """
        Creates a type instance with the given value

        :param value: value to build the type
        :return: The built type if the value is valid. None otherwise
        :rtype: IType or None
        """
        pass

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        """
        Get the list of opcodes to check if an value is of this type

        :return: A list of opcodes to check a value type
        """
        return [(Opcode.ISTYPE, self.stack_item)]

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        """
        Gets the class symbols of this type

        :return: a dictionary that maps each symbol in the class type with its name
        """
        return {}

    def union_type(self, other_type: IType) -> IType:
        """
        Gets a type that is an union of `self` and `other_type`

        :param other_type: type that'll be united with `self`
        :type other_type: IType
        :return: an union of this type and other_type
        :rtype: IType
        """
        from boa3.model.type.annotation.uniontype import UnionType
        return UnionType.build((self, other_type))

    def except_type(self, other_type: IType) -> IType:
        """
        Gets a type that is type of `self` but is not type of `other_type`.
        If `other_type` is an implementation of `self`, returns `self`

        :param other_type: type that won't be allowed in the resulting type
        :type other_type: IType
        :return: an union of this type and the exclusion of other_type
        :rtype: IType
        """
        return self

    def intersect_type(self, other_type: IType) -> IType:
        """
        Gets a type that is the intersection of `self` but is not type of `other_type`.
        If `other_type` is an implementation of `self`, returns `other_type`
        If `other_type` and `self` has no values in common, returns None

        :param other_type: type that'll be intersected with self
        :type other_type: IType
        :return: an intersection of this type and the exclusion of other_type
        :rtype: IType
        """
        if self.is_type_of(other_type):
            return other_type
        if other_type.is_type_of(self):
            return self

        from boa3.model.type.annotation.uniontype import UnionType
        if isinstance(other_type, UnionType):
            same_types = [x for x in other_type.union_types if self.is_type_of(x)]
            if len(same_types) == 0:
                return self
            elif len(same_types) == 1:
                return same_types[0]
            else:
                return UnionType.build(same_types)

        from boa3.model.type.type import Type
        return Type.none

    def is_equal(self, other: Any) -> bool:
        if not isinstance(other, IType):
            return False
        return self.is_type_of(other) and other.is_type_of(self)



from typing import Union

from boa3.builtin.type import ByteString


class StorageMap:
    """
    The key-value storage for the specific prefix in the given storage context.
    """

    def __init__(self, context, prefix: ByteString):
        from boa3.builtin.interop.storage.storagecontext import StorageContext

        self._context: StorageContext
        self._prefix: ByteString

    def get(self, key: ByteString) -> bytes:
        """
        Gets a value from the map based on the given key.

        :param key: value identifier in the store
        :type key: str or bytes
        :return: the value corresponding to given key for current storage context
        :rtype: bytes
        """
        pass

    def put(self, key: ByteString, value: Union[int, ByteString]):
        """
        Inserts a given value in the key-value format into the map.

        :param key: the identifier in the store for the new value
        :type key: str or bytes
        :param value: value to be stored
        :type value: int or str or bytes
        """
        pass

    def delete(self, key: ByteString):
        """
        Removes a given key from the map if exists.

        :param key: the identifier in the store for the new value
        :type key: str or bytes
        """
        pass



from typing import Dict

from boa3.model.builtin.interop.nativecontract import NeoContractMethod
from boa3.model.variable import Variable


class GetNextBlockValidatorsMethod(NeoContractMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.ecpointtype import ECPointType

        identifier = 'get_next_block_validators'
        native_identifier = 'getNextBlockValidators'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, native_identifier, args, return_type=Type.list.build([ECPointType.build()]))



from typing import Dict

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class GetRandomMethod(InteropMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'get_random'
        syscall = 'System.Runtime.GetRandom'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=Type.int)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError, CompilerWarning
from boa3.model.builtin.builtin import Builtin
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestException(BoaTest):
    default_folder: str = 'test_sc/exception_test'

    EXCEPTION_EMPTY_MESSAGE = String(Builtin.Exception.default_message).to_bytes()

    def test_raise_exception_empty_message(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # if arg0 < 0
            + Opcode.PUSH0
            + Opcode.LT
            + Opcode.JMPIFNOT
            + Integer(14).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHDATA1      # raise Exception
            + Integer(len(self.EXCEPTION_EMPTY_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + self.EXCEPTION_EMPTY_MESSAGE
            + Opcode.THROW
            + Opcode.RET
        )

        path = self.get_contract_path('RaiseExceptionEmptyMessage.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        self.run_smart_contract(engine, path, 'test_raise', 10)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.UNHANDLED_EXCEPTION_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'test_raise', -10)

    def test_raise_exception_with_message(self):
        exception_message = String('raised an exception').to_bytes()
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # if arg0 < 0
            + Opcode.PUSH0
            + Opcode.LT
            + Opcode.JMPIFNOT
            + Integer(24).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHDATA1      # raise Exception('raised an exception')
            + Integer(len(exception_message)).to_byte_array(signed=True, min_length=1)
            + exception_message
            + Opcode.THROW
            + Opcode.RET
        )

        path = self.get_contract_path('RaiseExceptionWithMessage.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        self.run_smart_contract(engine, path, 'test_raise', 10)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.UNHANDLED_EXCEPTION_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'test_raise', -10)

    def test_raise_exception_without_call(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # if arg0 < 0
            + Opcode.PUSH0
            + Opcode.LT
            + Opcode.JMPIFNOT
            + Integer(14).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHDATA1      # raise Exception
            + Integer(len(self.EXCEPTION_EMPTY_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + self.EXCEPTION_EMPTY_MESSAGE
            + Opcode.THROW
            + Opcode.RET
        )

        path = self.get_contract_path('RaiseExceptionWithoutCall.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        self.run_smart_contract(engine, path, 'test_raise', 10)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.UNHANDLED_EXCEPTION_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'test_raise', -10)

    def test_raise_variable_exception(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1  # x = Exception
            + Integer(len(self.EXCEPTION_EMPTY_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + self.EXCEPTION_EMPTY_MESSAGE
            + Opcode.STLOC0
            + Opcode.LDARG0     # if arg0 < 0
            + Opcode.PUSH0
            + Opcode.LT
            + Opcode.JMPIFNOT
            + Integer(4).to_byte_array(signed=True, min_length=1)
            + Opcode.LDLOC0         # raise x
            + Opcode.THROW
            + Opcode.RET
        )

        path = self.get_contract_path('RaiseVariableException.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        self.run_smart_contract(engine, path, 'test_raise', 10)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.UNHANDLED_EXCEPTION_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'test_raise', -10)

    def test_raise_exception_variable_message(self):
        message = 'raised an exception'
        exception_message = String(message).to_bytes()
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1  # x = 'raised an exception'
            + Integer(len(exception_message)).to_byte_array(signed=True, min_length=1)
            + exception_message
            + Opcode.STLOC0
            + Opcode.LDARG0     # if arg0 < 0
            + Opcode.PUSH0
            + Opcode.LT
            + Opcode.JMPIFNOT
            + Integer(24).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHDATA1      # raise Exception(x)
            + Integer(len(exception_message)).to_byte_array(signed=True, min_length=1)
            + exception_message
            + Opcode.THROW
            + Opcode.RET
        )

        path = self.get_contract_path('RaiseExceptionVariableMessage.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        self.run_smart_contract(engine, path, 'test_raise', 10)

        with self.assertRaisesRegex(TestExecutionException, message):
            self.run_smart_contract(engine, path, 'test_raise', -10)

    def test_raise_specific_exception(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # if arg0 < 0
            + Opcode.PUSH0
            + Opcode.LT
            + Opcode.JMPIFNOT
            + Integer(14).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHDATA1      # raise ValueError
            + Integer(len(self.EXCEPTION_EMPTY_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + self.EXCEPTION_EMPTY_MESSAGE
            + Opcode.THROW
            + Opcode.RET
        )

        path = self.get_contract_path('RaiseSpecificException.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        self.run_smart_contract(engine, path, 'test_raise', 10)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.UNHANDLED_EXCEPTION_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'test_raise', -10)

    def test_raise_mismatched_type(self):
        path = self.get_contract_path('RaiseMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_try_except_without_exception(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.TRY        # try:
            + Integer(7).to_byte_array(signed=True, min_length=1)  # jmp to exception
            + Integer(0).to_byte_array(signed=True, min_length=1)  # jmp to finally if exists
            + Opcode.LDARG0         # x = arg
            + Opcode.STLOC0
            + Opcode.JMP        # except:
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.PUSH0          # x = 0
            + Opcode.STLOC0
            + Opcode.ENDTRY
            + Integer(2).to_byte_array(signed=True, min_length=1)
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('TryExceptWithoutException.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'test_try_except', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'test_try_except', -110)
        self.assertEqual(-110, result)

    def test_try_except_base_exception(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.TRY        # try:
            + Integer(7).to_byte_array(signed=True, min_length=1)  # jmp to exception
            + Integer(0).to_byte_array(signed=True, min_length=1)  # jmp to finally if exists
            + Opcode.LDARG0         # x = arg
            + Opcode.STLOC0
            + Opcode.JMP        # except BaseException:
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.PUSH0          # x = 0
            + Opcode.STLOC0
            + Opcode.ENDTRY
            + Integer(2).to_byte_array(signed=True, min_length=1)
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('TryExceptBaseException.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'test_try_except', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'test_try_except', -110)
        self.assertEqual(-110, result)

    def test_try_except_specific_exception(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.TRY        # try:
            + Integer(7).to_byte_array(signed=True, min_length=1)  # jmp to exception
            + Integer(0).to_byte_array(signed=True, min_length=1)  # jmp to finally if exists
            + Opcode.LDARG0         # x = arg
            + Opcode.STLOC0
            + Opcode.JMP        # except ValueError:
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.PUSH0          # x = 0
            + Opcode.STLOC0
            + Opcode.ENDTRY
            + Integer(2).to_byte_array(signed=True, min_length=1)
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('TryExceptSpecificException.py')
        output = self.assertCompilerLogs(CompilerWarning.UsingSpecificException, path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'test_try_except', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'test_try_except', -110)
        self.assertEqual(-110, result)

    def test_try_except_with_name(self):
        path = self.get_contract_path('TryExceptWithName.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_try_except_finally(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.PUSH4
            + Opcode.DIV
            + Opcode.STLOC0
            + Opcode.TRY        # try:
            + Integer(9).to_byte_array(signed=True, min_length=1)   # jmp to exception
            + Integer(15).to_byte_array(signed=True, min_length=1)  # jmp to finally if exists
            + Opcode.LDLOC0         # x += arg
            + Opcode.LDARG0
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.JMP        # except ValueError:
            + Integer(6).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.LDLOC0         # x = -x
            + Opcode.NEGATE
            + Opcode.STLOC0
            + Opcode.ENDTRY
            + Integer(7).to_byte_array(signed=True, min_length=1)
            + Opcode.LDLOC0     # finally
            + Opcode.PUSH2          # x *= 2
            + Opcode.MUL
            + Opcode.STLOC0
            + Opcode.ENDFINALLY
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('TryExceptFinally.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'test_try_except', 10)
        self.assertEqual(24, result)
        result = self.run_smart_contract(engine, path, 'test_try_except', -110)
        self.assertEqual(-274, result)

    def test_try_except_else(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.TRY        # try:
            + Integer(7).to_byte_array(signed=True, min_length=1)  # jmp to exception
            + Integer(0).to_byte_array(signed=True, min_length=1)  # jmp to finally if exists
            + Opcode.LDARG0         # x = arg
            + Opcode.STLOC0
            + Opcode.JMP        # except BaseException:
            + Integer(7).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.PUSH0          # x = 0
            + Opcode.STLOC0
            + Opcode.JMP        # else:
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.LDARG0         # x = -arg
            + Opcode.NEGATE
            + Opcode.STLOC0
            + Opcode.ENDTRY
            + Integer(2).to_byte_array(signed=True, min_length=1)
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('TryExceptElse.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'test_try_except', 10)
        self.assertEqual(-10, result)
        result = self.run_smart_contract(engine, path, 'test_try_except', -110)
        self.assertEqual(110, result)

    def test_try_except_else_finally(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.PUSH4
            + Opcode.DIV
            + Opcode.STLOC0
            + Opcode.TRY        # try:
            + Integer(9).to_byte_array(signed=True, min_length=1)   # jmp to exception
            + Integer(21).to_byte_array(signed=True, min_length=1)  # jmp to finally if exists
            + Opcode.LDLOC0         # x += arg
            + Opcode.LDARG0
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.JMP        # except ValueError:
            + Integer(8).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.LDLOC0         # x = -x
            + Opcode.NEGATE
            + Opcode.STLOC0
            + Opcode.JMP        # else:
            + Integer(6).to_byte_array(signed=True, min_length=1)
            + Opcode.LDLOC0         # x += arg
            + Opcode.LDARG0
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.ENDTRY
            + Integer(7).to_byte_array(signed=True, min_length=1)
            + Opcode.LDLOC0     # finally
            + Opcode.PUSH2          # x *= 2
            + Opcode.MUL
            + Opcode.STLOC0
            + Opcode.ENDFINALLY
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('TryExceptElseFinally.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'test_try_except', 10)
        self.assertEqual(44, result)
        result = self.run_smart_contract(engine, path, 'test_try_except', -110)
        self.assertEqual(-494, result)



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.mutable.listtype import ListType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class CopyMethod(IBuiltinMethod):
    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type
        identifier = 'copy'

        self._allowed_types = [Type.dict, Type.list]
        default_type = Type.list
        if not self._is_valid_type(arg_value):
            arg_value = default_type

        args: Dict[str, Variable] = {'self': Variable(arg_value)}
        super().__init__(identifier, args, return_type=arg_value)

    def _is_valid_type(self, arg_type: Optional[IType]) -> bool:
        return (isinstance(arg_type, IType) and
                any(allowed_type.is_type_of(arg_type) for allowed_type in self._allowed_types))

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if self._arg_self.type is Type.list:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_self.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        return isinstance(params[0], IExpression) and isinstance(params[0].type, ListType)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [
            (Opcode.UNPACK, b''),
            (Opcode.PACK, b''),
        ]

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and len(value) == 1:
            value = value[0]
        if value == self.args['self']:
            return self

        from boa3.model.builtin.classmethod.copydictmethod import CopyDictMethod
        from boa3.model.builtin.classmethod.copylistmethod import CopyListMethod
        from boa3.model.type.type import Type

        if Type.dict.is_type_of(value):
            return CopyDictMethod(value)

        return CopyListMethod(value)



import ast

from boa3.analyser.astanalyser import IAstAnalyser


class ConstructAnalyser(IAstAnalyser, ast.NodeTransformer):
    """
    This class is responsible for pre processing Python constructs

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.
    """

    def __init__(self, ast_tree: ast.AST, log: bool = False):
        super().__init__(ast_tree, log=log)
        self.visit(self._tree)

    @property
    def tree(self) -> ast.AST:
        """
        Gets the analysed abstract syntax tree

        :return: the analysed ast
        """
        return self._tree

    def visit_Call(self, call: ast.Call) -> ast.AST:
        """
        Visitor of a function call node

        :param call: the python ast function call node
        """
        if isinstance(call.func, ast.Attribute):
            from boa3.model.builtin.builtin import Builtin
            if call.func.attr == Builtin.ScriptHash.identifier:
                from boa3.constants import SYS_VERSION_INFO
                from boa3.model.type.type import Type
                types = {
                    Type.int.identifier: int,
                    Type.str.identifier: str,
                    Type.bytes.identifier: bytes
                }
                literal: tuple = ((ast.Constant,)
                                  if SYS_VERSION_INFO >= (3, 8)
                                  else (ast.Num, ast.Str, ast.Bytes))

                if isinstance(call.func.value, literal) and len(call.args) == 0:
                    value = ast.literal_eval(call.func.value)
                    if not isinstance(value, tuple(types.values())):
                        return call
                elif (isinstance(call.func.value, ast.Name)  # checks if is the name of a type
                      and call.func.value.id in types  # and if the arguments is from the same type
                      and len(call.args) == 1
                      and isinstance(call.args[0], literal)):
                    value = ast.literal_eval(call.args[0])
                    if not isinstance(value, (types[call.func.value.id],)):
                        return call
                else:
                    return call

                from boa3.neo import to_script_hash
                # value must be bytes
                if isinstance(value, int):
                    from boa3.neo.vm.type.Integer import Integer
                    value = Integer(value).to_byte_array()
                elif isinstance(value, str):
                    from boa3.neo.vm.type.String import String
                    value = String(value).to_bytes()
                return self.parse_to_node(str(to_script_hash(value)), call)

        return call



import ast
from abc import abstractmethod
from typing import Optional

from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType


class IExpression(ISymbol):
    """
    An interface used to represent expressions
    """

    def __init__(self, origin_node: Optional[ast.AST] = None):
        self._origin_node = origin_node

    @property
    @abstractmethod
    def type(self) -> IType:
        """
        Gets the type of the evaluated expression

        :return: the resulting type when the expression is evaluated
        """
        pass

    @property
    def origin(self) -> ast.AST:
        """
        Returns the method origin ast node.

        :return: the ast node that describes this method. None if it is not from a ast.
        """
        return self._origin_node



from typing import List

from boa3.builtin.compile_time import public
from boa3.builtin.interop.runtime import Notification
from boa3.builtin.type import UInt160
from boa3_test.test_sc.interop_test.runtime.GetNotifications import with_param


@public
def main(args: list, key: UInt160) -> List[Notification]:
    return with_param(args, key)



from __future__ import annotations

import base64
from typing import Any, Dict, Optional

from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from test_runner.blockchain.contract import TestRunnerContract
from test_runner.blockchain.contractcollection import ContractCollection


class TestRunnerStorage:
    _storage_contract_name_key = 'name'
    _storage_contract_hash_key = 'hash'
    _storage_contract_values_key = 'values'
    _storage_key_key = 'key'
    _storage_item_key = 'value'

    def __init__(self, contract: TestRunnerContract, values: Dict[StorageKey, StorageItem]):
        self._contract: TestRunnerContract = contract
        self._values: Dict[StorageKey, StorageItem] = values

    @property
    def contract(self) -> Optional[TestRunnerContract]:
        if isinstance(self._contract, TestRunnerContract):
            return self._contract
        return None

    @property
    def values(self) -> Dict[StorageKey, StorageItem]:
        return self._values

    @classmethod
    def from_json(cls, json: Dict[str, Any], contracts: ContractCollection = None) -> TestRunnerStorage:
        keys = set(json.keys())
        if not keys.issubset([cls._storage_contract_name_key,
                              cls._storage_contract_hash_key,
                              cls._storage_contract_values_key]):
            return None

        contract_name = json[cls._storage_contract_name_key]
        contract_hash = json[cls._storage_contract_hash_key]
        if contract_hash in contracts:
            contract = contracts[contract_hash]
        elif contract_name in contracts:
            contract = contracts[contract_name]
        else:
            contract = None

        storage = {}
        storage_required_fields = [cls._storage_key_key, cls._storage_item_key]
        for item_json in json[cls._storage_contract_values_key]:
            if not set(item_json.keys()).issubset(storage_required_fields):
                continue

            key = StorageKey.from_json(item_json[cls._storage_key_key])
            item = StorageItem.from_json(item_json[cls._storage_item_key])

            storage[key] = item

        return cls(contract, storage)

    def is_valid_identifier(self, item: str) -> bool:
        if isinstance(self._contract, TestRunnerContract):
            return self._contract.is_valid_identifier(item)

        return False

    def __str__(self) -> str:
        return str(self._values)

    def __repr__(self) -> str:
        return self._values.__repr__()


class StorageKey:
    def __init__(self, key: bytes):
        self._key: bytes = key

    @classmethod
    def from_json(cls, json: str) -> StorageKey:
        decoded: bytes = base64.b64decode(json)
        return cls(decoded)

    def as_str(self) -> str:
        return String.from_bytes(self._key)

    def as_bytes(self) -> bytes:
        return self._key

    def __str__(self):
        return self.as_str()

    def __repr__(self):
        return str(self)

    def __hash__(self) -> int:
        return self._key.__hash__()

    def __eq__(self, other) -> bool:
        return (other == self.as_bytes()
                or other == self.as_str()
                or (isinstance(other, StorageKey) and self._key == other._key))


class StorageItem:
    def __init__(self, value: bytes):
        self._value: bytes = value

    @classmethod
    def from_json(cls, json: str) -> StorageItem:
        decoded: bytes = base64.b64decode(json)
        return cls(decoded)

    def as_bytes(self) -> bytes:
        return self._value

    def as_str(self) -> str:
        return String.from_bytes(self._value)

    def as_int(self) -> int:
        return Integer.from_bytes(self._value)

    def __str__(self) -> str:
        return self._value.__str__()

    def __repr__(self):
        return str(self)

    def __hash__(self) -> int:
        return self._value.__hash__()

    def __eq__(self, other) -> bool:
        return (other == self.as_bytes()
                or other == self.as_int()
                or other == self.as_str()
                or (isinstance(other, StorageItem) and self._value == other._value))
